var fs = require("fs");
var path = require("path");
var Promise = require('promise');
var SlJsInfra = require("sl-js-infra").SlJsInfra;
var ValidationUtils = SlJsInfra.Utils.ValidationUtils;
var mkdirp = require('mkdirp');
/**
 * Map between uniqueId in generated file to relevant uniqueId in source file.
 * This object will be added to 'window' object due to instrumentation, in order to resolve uniquId's in browser agent
 * @constructor
 */

var BLOB_ID = "sl-mapping";
function SlMapper(outputPath, backendProxy, buildSessionId, logger) {
    ValidationUtils.verifyNotNullOrEmpty(backendProxy, "backendProxy");
    ValidationUtils.verifyNotNullOrEmpty(buildSessionId, "buildSessionId");
    ValidationUtils.verifyNotNullOrEmpty(logger, "logger");
    this.mapping = {};
    this.outputPath = outputPath;
    this.backendProxy = backendProxy;
    this.buildSessionId = buildSessionId;
    this.logger = logger;

    this.mapMethod = function(method) {
        var key = this.createUniqueIdKey(method, false);
        if(key) {
            this.mapping[key] = method.uniqueId;
        }
    }
    this.mapBranch = function(branch) {
        var key = this.createUniqueIdKey(branch, true);
        if(key) {
            this.mapping[key] = branch.uniqueId;
        }
    }


    this.getFilePath = function() {
        return path.resolve(this.outputPath);
    }

    this.writeToFile = function () {
        var filePath = this.getFilePath();
        var outputFolder = path.dirname(filePath);
        mkdirp.sync(outputFolder);
        var fileContent = this.createFileContent();
        try {
            fs.writeFileSync(filePath, fileContent);
        }catch (e) {
            this.logger.error("Failed to write sl-mapping file Error: ", e)
        }
    }

    this.postToServer = function () {
        var context = this;
        return new Promise(function (resolve , reject) {
            context.backendProxy.submitBlob(context.mapping, context.buildSessionId, BLOB_ID, function (err)  {
                if (err) {
                    context.logger.error("Error while submitting sl-mapping, '%s'", err);
                    reject(err)
                }  else {
                    // TODO: SLDEV-4974
                    context.logger.info("sl-mapping submitted successfully");
                    resolve();
                }
            });
        });


    }

    this.createUniqueIdKey = function(element, isBranch){
        var delimiter = isBranch ? "|" : "@";
        // Istanbul report branches with parent location, to align the key for branches based on parent position.
        var position = isBranch ? element.generatedData.parentPosition : element.generatedData.position;
        var key = [element.generatedData.absoluteFilename, position.join(",")].join(delimiter);
        if(isBranch){
            key += "|" + element.index;
        }
        return key;
    }


    this.createFileContent = function () {
        var content ="var slMapping =" + JSON.stringify(this.mapping, null, "\t") + "\n" +
            "        if(window && (!window.$Sealights || !window.$Sealights.slMapping)){\n" +
            "            window.$Sealights = window.$Sealights  || {};\n" +
            "            window.$Sealights.slMapping = slMapping;\n" +
            "        }";
        return content;
    }

    function validateParams() {
        for(var i = 0; i< arguments.length; i++){
            if(!arguments[i]){
                return false;
            }
            return true;
        }
    }
}
module.exports.SlMapper = SlMapper;







